/*
Copyright 2012-2017 TacDev [http://www.tacdev.eu] Contact: support@tacdev.eu

This software is distributed under the MIT license

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

using System;
using System.IO;
using System.Security.Cryptography;
using System.Text;

namespace TacDev.Shared
{
    public static class Encryptor
    {
        #region Settings
        private static RNGCryptoServiceProvider rand = new RNGCryptoServiceProvider();
        private static int KeySize = 256; // Increase size as required for your safety concerns
        private static int PasswordLength = 16; // This is not really safe but a tradeoff vs. UDP size
        private static int IVLength = 16; // Increase size as required for your safety concerns
        private static string Hash = "SHA1";
        private static byte[] Salt = new byte[] { 0x42, 0xe2, 0x17, 0x41, 0xA8, 0x12, 0x27, 0x19, 0x15, 0xc3, 0xd3, 0x53, 0xa1, 0xA3, 0x52, 0xf1 };
        #endregion

        /// <summary>
        /// The symmetrical part of the encryption, that returns a password and IV that then can be
        /// encrypted using asymmetrical encryption
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="value"></param>
        /// <param name="password"></param>
        /// <param name="IV"></param>
        /// <returns></returns>
        private static byte[] EncryptSymmetric<T>(byte[] value, out byte[] password, out byte[] IV) where T : SymmetricAlgorithm, new()
        {
            // Random IV
            IV = new byte[IVLength];
            rand.GetBytes(IV);

            // Random Password
            password = new byte[PasswordLength]; // Sorry UDP size restrictions. Its not a nuclear factory.
            rand.GetBytes(password);

            byte[] encrypted;

            using (T cipher = new T())
            {
                // From the password we derive a key with the correct length.... the password can be much shorter than the key length
                // The salt is fixed and does not get transmitted. The attacker can get it from this application though.
                PasswordDeriveBytes _passwordBytes = new PasswordDeriveBytes(password, Salt, Hash, 2);
                byte[] keyBytes = _passwordBytes.GetBytes(KeySize / 8);

                cipher.Mode = CipherMode.CBC;

                // Send to encryption stream
                using (ICryptoTransform encryptor = cipher.CreateEncryptor(keyBytes, IV))
                {
                    using (MemoryStream to = new MemoryStream())
                    {
                        using (CryptoStream writer = new CryptoStream(to, encryptor, CryptoStreamMode.Write))
                        {
                            writer.Write(value, 0, value.Length);
                            writer.FlushFinalBlock();

                            encrypted = to.ToArray();
                        }
                    }
                }
                cipher.Clear();
            }

            return encrypted;
        }

        /// <summary>
        /// The symmetrical part of the decryption, that accepts a password and IV received
        /// from the asymmetrical decryption
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="value"></param>
        /// <param name="password"></param>
        /// <param name="IV"></param>
        /// <returns></returns>
        private static byte[] DecryptSymmetric<T>(byte[] value, byte[] password, byte[] IV) where T : SymmetricAlgorithm, new()
        {
            byte[] decrypted;
            byte[] decMsg;
            int decryptedByteCount = 0;

            using (T cipher = new T())
            {
                // From the password we derive a key with the correct length.... the password can be much shorter than the key length
                // The salt is fixed and does not get transmitted. The attacker can get it from this application though.
                PasswordDeriveBytes _passwordBytes = new PasswordDeriveBytes(password, Salt, Hash, 2);
                byte[] keyBytes = _passwordBytes.GetBytes(KeySize / 8);

                cipher.Mode = CipherMode.CBC;

                // Send to encryption stream
                try
                {
                    using (ICryptoTransform decryptor = cipher.CreateDecryptor(keyBytes, IV))
                    {
                        using (MemoryStream from = new MemoryStream(value))
                        {
                            using (CryptoStream reader = new CryptoStream(from, decryptor, CryptoStreamMode.Read))
                            {
                                decrypted = new byte[value.Length];
                                decryptedByteCount = reader.Read(decrypted, 0, decrypted.Length);
                            }
                        }
                    }
                }
                catch (Exception)
                {
                    return null;
                }

                decMsg = new byte[decryptedByteCount];
                Buffer.BlockCopy(decrypted, 0, decMsg, 0, decryptedByteCount);

                cipher.Clear();
            }

            return decMsg;
        }

        /// <summary>
        /// This is what you call to encrypt a payload. It applies a symmetric encryption to the message and encrypts the password
        /// used for that encryption with the RSA asymmetric cypher
        /// </summary>
        /// <param name="message"></param>
        /// <param name="SenderKey"></param>
        /// <param name="ReceiverKey"></param>
        /// <returns></returns>
        public static byte[] EncryptPayload(byte[] message, RSACryptoServiceProvider SenderKey, RSACryptoServiceProvider ReceiverKey)
        {
            // We sign the data so the receiver can verify it is from us. The signature is inside the encrypted container.
            byte[] signature = SenderKey.SignData(message, HashAlgorithm.Create("SHA256"));

            // The complete container to be encrypted has the signature and the actual message
            byte[] complete = null;

            // The size of the signature depends on the key size so we store its length to get it back out
            using (MemoryStream ms = new MemoryStream())
            {
                using (BinaryWriter binWriter = new BinaryWriter(ms))
                {
                    binWriter.Write(signature.Length);
                    binWriter.Write(signature);
                    binWriter.Write(message.Length);
                    binWriter.Write(message);
                    complete = ms.ToArray();
                }
            }

            // Max length of encryption = ReceiverKey.KeySize / 8 - 2 - (2 * SHA1 Hash Size = 20)

            // We encrypt with a symmetric cypher and get back a random password and IV that have to be encrypted with the asymmetric cypher
            byte[] RandomPasswordFromAES;
            byte[] RandomIVFromAES;
            byte[] encPayload = EncryptSymmetric<AesManaged>(complete, out RandomPasswordFromAES, out RandomIVFromAES);

            // We assemble everything that has to be asymmetrically encrypted.... the password and the IV of the symmetric encryption
            complete = new byte[RandomPasswordFromAES.Length + RandomIVFromAES.Length];
            Buffer.BlockCopy(RandomPasswordFromAES, 0, complete, 0, RandomPasswordFromAES.Length);
            Buffer.BlockCopy(RandomIVFromAES, 0, complete, RandomPasswordFromAES.Length, RandomIVFromAES.Length);

            // The actual asymmetrical encryption
            byte[] encAESKey = ReceiverKey.Encrypt(complete, true);

            // Assemble encrypted key and encrypted container into byte array
            using (MemoryStream ms = new MemoryStream())
            {
                using (BinaryWriter binWriter = new BinaryWriter(ms))
                {
                    binWriter.Write(encPayload.Length);
                    binWriter.Write(encPayload);
                    binWriter.Write(encAESKey.Length);
                    binWriter.Write(encAESKey);
                    complete = ms.ToArray();
                }
            }

            return complete;
        }

        /// <summary>
        /// This is what you call to decrypt a payload. It decrypts the password used for the symmetric encryption using the RSA
        /// asymmetric cypher and then decrypts the message and signature with the the symmetric decryption routine
        /// </summary>
        /// <param name="message"></param>
        /// <param name="SenderKey"></param>
        /// <param name="ReceiverKey"></param>
        /// <returns></returns>
        public static byte[] DecryptPayload(byte[] message, RSACryptoServiceProvider SenderKey, RSACryptoServiceProvider ReceiverKey)
        {
            // Read the (symmetric) encrypted payload and the (asymmetric) encrypted key from the byte array
            byte[] encPayload = null;
            byte[] encAESKey = null;

            using (MemoryStream ms = new MemoryStream(message))
            {
                using (BinaryReader binReader = new BinaryReader(ms))
                {
                    int num = binReader.ReadInt32();
                    encPayload = binReader.ReadBytes(num);
                    num = binReader.ReadInt32();
                    encAESKey = binReader.ReadBytes(num);
                }
            }

            // We decrypt the Password and the IV for the AES container
            byte[] decAESKeyAndIV = ReceiverKey.Decrypt(encAESKey, true);

            // We take the two apart
            byte[] Password = new byte[PasswordLength];
            byte[] IV = new byte[IVLength];
            Buffer.BlockCopy(decAESKeyAndIV, 0, Password, 0, Password.Length);
            Buffer.BlockCopy(decAESKeyAndIV, Password.Length, IV, 0, IV.Length);

            // We decrypt the AES container with the decrypted password and IV
            byte[] decPayload = DecryptSymmetric<AesManaged>(encPayload, Password, IV);

            // We take apart the signature and the message
            byte[] decSignature = null;
            byte[] decMessage = null;

            using (MemoryStream ms = new MemoryStream(decPayload))
            {
                using (BinaryReader binReader = new BinaryReader(ms))
                {
                    int num = binReader.ReadInt32();
                    decSignature = binReader.ReadBytes(num);
                    num = binReader.ReadInt32();
                    decMessage = binReader.ReadBytes(num);
                }
            }

            // We use the senders key to verify that the message was not altered
            if (!SenderKey.VerifyData(decMessage, HashAlgorithm.Create("SHA256"), decSignature))
                return null;

            return decMessage;
        }

    }
}